--[[
Interface: 1.4.1.0 b5334

Copyright (C) GtX (Andy), 2019

Author: GtX | Andy
Date: 02.05.2019
Version: 1.0.0.0

Contact:
https://forum.giants-software.com
https://github.com/GtX-Andy

Script Idea: GtX and ccs101

History:
V 1.0.0.0 @ 02.05.2019 - Release Version

Important:
No changes are to be made to this script without permission from GtX | Andy
An diesem Skript dürfen ohne Genehmigung von GtX | Andy keine Änderungen vorgenommen werden
]]


IndividualSiloPlaceable = {}

local IndividualSiloPlaceable_mt = Class(IndividualSiloPlaceable, Placeable)
InitObjectClass(IndividualSiloPlaceable, "IndividualSiloPlaceable")

IndividualSiloPlaceable.PRICE_SELL_FACTOR = 0.7
IndividualSiloPlaceable.LOWEST_SELL_PRICE = 0.5

IndividualSiloPlaceable.DISPLAY_TYPES = {
    ["FILL_LEVEL"] = true,
    ["PERCENT"] = true,
    ["CAPACITY"] = true
}

function IndividualSiloPlaceable:new(isServer, isClient, customMt)
    local self = Placeable:new(isServer, isClient, customMt or IndividualSiloPlaceable_mt)

    self.vehiclesInRange = {}
    self.playerInRange = false
    self.numVehiclesInRange = 0

    self.totalFillTypeSellPrice = 0

    self.fillLevelTextBackup = g_i18n:getText("info_fillLevel")

    registerObjectClassName(self, "IndividualSiloPlaceable")

    return self
end

function IndividualSiloPlaceable:load(xmlFilename, x,y,z, rx,ry,rz, initRandom)
    if not IndividualSiloPlaceable:superClass().load(self, xmlFilename, x,y,z, rx,ry,rz, initRandom) then
        return false
    end

    local xmlFile = loadXMLFile("TempXML", xmlFilename)

    local key = "placeable.individualSilo"
    local storageKey = key .. ".storage"

    self.storagePerFarm = false -- Not an option for this silo type.
    self.foreignSilo = Utils.getNoNil(getXMLBool(xmlFile, key .. "#foreignSilo"), self.storagePerFarm)

    self.unloadingStation = UnloadingStation:new(self.isServer, self.isClient)
    self.unloadingStation:load(self.nodeId, xmlFile, key .. ".unloadingStation")
    self.unloadingStation.owningPlaceable = self
    self.unloadingStation.hasStoragePerFarm = self.storagePerFarm
    self.unloadingStation.storageRadius = 0

    self.loadingStation = LoadingStation:new(self.isServer, self.isClient)
    self.loadingStation:load(self.nodeId, xmlFile, key .. ".loadingStation")
    self.loadingStation.owningPlaceable = self
    self.loadingStation.hasStoragePerFarm = self.storagePerFarm
    self.loadingStation.storageRadius = 0

    self.loadingStation.getAllFillLevels = Utils.overwrittenFunction(self.loadingStation.getAllFillLevels, function(self, superFunc, farmId)
        for _, sourceStorage in pairs(self.sourceStorages) do
            local fillTypeIndex = sourceStorage.currentStoredFillType
            if fillTypeIndex ~= nil and sourceStorage.fillLevels ~= nil then
                local fillLevel = Utils.getNoNil(sourceStorage.fillLevels[fillTypeIndex], 0)
                return {[fillTypeIndex] = fillLevel}, sourceStorage:getCapacity()
            end
        end

        return superFunc(self, farmId)
    end)

    local storageNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, storageKey .. "#node"))
    if storageNode ~= nil then
        local storage = Storage:new(self.isServer, self.isClient)
        if storage:load(storageNode, xmlFile, storageKey) then
            self.storage = storage
            self.storage.storageOwner = self
            self.storage.currentStoredFillType = nil
            self.storage.foreignSilo = self.foreignSilo

            self.storage.setFillLevel = Utils.overwrittenFunction(self.storage.setFillLevel, function(self, superFunc, fillLevel, fillTypeIndex)
                fillLevel = MathUtil.clamp(fillLevel, 0, self.capacityPerFillType)

                if self.fillLevels[fillTypeIndex] ~= nil and fillLevel ~= self.fillLevels[fillTypeIndex] then
                    self.fillLevels[fillTypeIndex] = fillLevel

                    if fillLevel > 0.001 then
                        if fillTypeIndex ~= self.currentStoredFillType then
                            self.currentStoredFillType = fillTypeIndex
                        end
                    else
                        if self.currentStoredFillType ~= nil then
                            self.currentStoredFillType = nil
                        end
                    end

                    if self.storageOwner ~= nil then
                        local capacity = self.capacityPerFillType
                        self.storageOwner:fillLevelUpdate(fillLevel, capacity, fillTypeIndex)
                    end

                    if self.isServer then
                        self:raiseDirtyFlags(self.storageDirtyFlag)
                    end
                end
            end)

            self.storage.getIsFillTypeSupported = Utils.overwrittenFunction(self.storage.getIsFillTypeSupported, function(self, superFunc, fillTypeIndex)
                if self.currentStoredFillType ~= nil then
                    return self.currentStoredFillType == fillTypeIndex
                end

                return self.fillTypes[fillTypeIndex] == true
            end)

            local i = 0
            while true do
                local displayKey = string.format("%s.displays.display(%d)", storageKey, i)
                if not hasXMLProperty(xmlFile, displayKey) then
                    break
                end

                local display = {}
                local displayType = Utils.getNoNil(getXMLString(xmlFile, displayKey .. "#type"), "FILL_LEVEL")
                if displayType ~= nil then
                    display.type = displayType:upper()

                    if IndividualSiloPlaceable.DISPLAY_TYPES[display.type] then
                        local numbers = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, displayKey .. "#node"))
                        if numbers ~= nil then
                            local numMeshes = getNumOfChildren(numbers)
                            if numMeshes > 0 then
                                display.numbers = numbers
                                display.showOnEmpty = Utils.getNoNil(getXMLBool(xmlFile, displayKey .. "#showOnEmpty"), true)
                                display.maxValue = (10 ^ (numMeshes)) - 1

                                if display.type == "CAPACITY" then
                                    self:setDisplayValue(display, self.storage.capacityPerFillType)
                                else
                                    self:setDisplayValue(display, 0)
                                end

                                if self.displays == nil then
                                    self.displays = {}
                                end

                                table.insert(self.displays, display)
                            else
                                g_logManager:xmlWarning(xmlFilename, "No mesh children found at node! '%s'", displayKey)
                            end
                        end
                    else
                        g_logManager:xmlWarning(xmlFilename, "Display type '%s' is not supported! Use FILL_LEVEL or PERCENT or CAPACITY instead.", displayType)
                    end
                end

                i = i + 1
            end

            i = 0
            while true do
                local iconKey = string.format("%s.icons.icon(%d)", storageKey, i)
                if not hasXMLProperty(xmlFile, iconKey) then
                    break
                end

                local icon = {}
                icon.node = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, iconKey .. "#node"))
                if icon.node ~= nil then
                    icon.color = StringUtil.getVectorNFromString(Utils.getNoNil(getXMLString(xmlFile, iconKey .. "#iconColor"), "1 1 1 1"), 4)

                    if self.fillTypeIcons == nil then
                        local originalMaterial = getMaterial(icon.node, 0)
                        local alwaysVisible = Utils.getNoNil(getXMLBool(xmlFile, storageKey .. ".icons#alwaysVisible"), false)

                        self.fillTypeIcons = {originalMaterial = originalMaterial, alwaysVisible = alwaysVisible, icons = {}}
                    end

                    setVisibility(icon.node, self.fillTypeIcons.alwaysVisible)

                    table.insert(self.fillTypeIcons.icons, icon)
                else
                    g_logManager:xmlWarning(xmlFilename, "No node given for '%s'", iconKey)
                end

                i = i + 1
            end

            i = 0
            while true do
                local fillPlaneKey = string.format("%s.fillPlanes.fillPlane(%d)", storageKey, i)
                if not hasXMLProperty(xmlFile, fillPlaneKey) then
                    break
                end

                local fillPlaneNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, fillPlaneKey .. "#node"))
                if fillPlaneNode ~= nil then
                    if self.fillPlanes == nil then
                        self.fillPlanes = {}
                    end

                    local fillPlane = {}

                    local minY, maxY = StringUtil.getVectorFromString(getXMLString(xmlFile, fillPlaneKey .. "#minMaxY"))
                    if minY ~= nil and maxY ~= nil then
                        fillPlane.minY = minY
                        fillPlane.maxY = maxY
                        fillPlane.updateTranslation = true
                    else
                        fillPlane.minY = 0.0
                        fillPlane.maxY = 0.0
                        fillPlane.updateTranslation = false
                    end

                    fillPlane.node = fillPlaneNode
                    fillPlane.originalMaterial = getMaterial(fillPlaneNode, 0)

                    fillPlane.alwaysVisible = Utils.getNoNil(getXMLBool(xmlFile, fillPlaneKey .. "#alwaysVisible"), false)

                    local j = 0
                    while true do
                        local materialKey = string.format("%s.materials.material(%d)", fillPlaneKey, j)
                        if not hasXMLProperty(xmlFile, materialKey) then
                            break
                        end

                        local fillTypeName = getXMLString(xmlFile, materialKey .. "#fillType")
                        if fillTypeName ~= nil then
                            local fillTypeIndex = g_fillTypeManager:getFillTypeIndexByName(fillTypeName)
                            if fillTypeIndex ~= nil then
                                local refMaterialNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, materialKey .. "#refNode"))
                                if refMaterialNode ~= nil and getHasClassId(refMaterialNode, ClassIds.SHAPE) then
                                    if fillPlane.materials == nil then
                                        fillPlane.materials = {}
                                    end

                                    if fillPlane.materials[fillTypeIndex] == nil then
                                        fillPlane.materials[fillTypeIndex] = getMaterial(refMaterialNode, 0)
                                    else
                                        g_logManager:xmlWarning(xmlFilename, "Duplicate material for %s at '%s'!", fillTypeName, materialKey)
                                    end
                                else
                                    g_logManager:xmlWarning(xmlFilename, "No reference material or node is not a shape for '%s'!", materialKey)
                                end
                            else
                                g_logManager:xmlWarning(xmlFilename, "Invalid fillType '%s' for '%s'!", fillTypeName, materialKey)
                            end
                        end

                        j = j + 1
                    end

                    setVisibility(fillPlane.node, fillPlane.alwaysVisible)
                    table.insert(self.fillPlanes, fillPlane)
                end

                i = i + 1
            end

            self.storages = {storage}
            self.interactionTriggerNode = I3DUtil.indexToObject(self.nodeId, getXMLString(xmlFile, key .. ".interactionTrigger#node"))
        else
            g_logManager:xmlWarning(xmlFilename, "Storage failed to load!", storageKey)
            return false
        end
    else
        g_logManager:xmlWarning(xmlFilename, "Missing 'node' for storage '%s'!", storageKey)
        return false
    end

    delete(xmlFile)

    return true
end

function IndividualSiloPlaceable:finalizePlacement()
    IndividualSiloPlaceable:superClass().finalizePlacement(self)

    if self.interactionTriggerNode ~= nil then
        addTrigger(self.interactionTriggerNode, "interactionTriggerCallback", self)
    end

    local storageSystem = g_currentMission.storageSystem

    self.unloadingStation:register(true)
    storageSystem:addUnloadingStation(self.unloadingStation)

    self.loadingStation:register(true)
    storageSystem:addLoadingStation(self.loadingStation)

    self.storage:setOwnerFarmId(self:getOwnerFarmId(), true)
    storageSystem:addStorage(self.storage)
    self.storage:register(true)

    storageSystem:addStorageToUnloadingStations(self.storage, {self.unloadingStation})
    storageSystem:addStorageToLoadingStations(self.storage, {self.loadingStation})
end

function IndividualSiloPlaceable:delete()
    if self.interactionTriggerNode ~= nil then
        removeTrigger(self.interactionTriggerNode)
    end

    local storageSystem = g_currentMission.storageSystem

    if self.unloadingStation ~= nil then
        storageSystem:removeStorageFromUnloadingStations(self.storage, {self.unloadingStation})

        storageSystem:removeUnloadingStation(self.unloadingStation)
        self.unloadingStation:delete()
        self.unloadingStation = nil
    end

    if self.loadingStation ~= nil then
        storageSystem:removeStorageFromLoadingStations(self.storage, {self.loadingStation})

        storageSystem:removeLoadingStation(self.loadingStation)
        self.loadingStation:delete()
        self.loadingStation = nil
    end

    storageSystem:removeStorage(self.storage)
    self.storage:delete()
    self.storages = nil

    unregisterObjectClassName(self)
    IndividualSiloPlaceable:superClass().delete(self)
end

function IndividualSiloPlaceable:readStream(streamId, connection)
    IndividualSiloPlaceable:superClass().readStream(self, streamId, connection)

    if connection:getIsServer() then
        local unloadingStationId = NetworkUtil.readNodeObjectId(streamId)
        self.unloadingStation:readStream(streamId, connection)
        g_client:finishRegisterObject(self.unloadingStation, unloadingStationId)

        local loadingStationId = NetworkUtil.readNodeObjectId(streamId)
        self.loadingStation:readStream(streamId, connection)
        g_client:finishRegisterObject(self.loadingStation, loadingStationId)

        local storageId = NetworkUtil.readNodeObjectId(streamId)
        self.storage:readStream(streamId, connection)
        g_client:finishRegisterObject(self.storage, storageId)
    end
end

function IndividualSiloPlaceable:writeStream(streamId, connection)
    IndividualSiloPlaceable:superClass().writeStream(self, streamId, connection)

    if not connection:getIsServer() then
        NetworkUtil.writeNodeObjectId(streamId, NetworkUtil.getObjectId(self.unloadingStation))
        self.unloadingStation:writeStream(streamId, connection)
        g_server:registerObjectInStream(connection, self.unloadingStation)

        NetworkUtil.writeNodeObjectId(streamId, NetworkUtil.getObjectId(self.loadingStation))
        self.loadingStation:writeStream(streamId, connection)
        g_server:registerObjectInStream(connection, self.loadingStation)

        NetworkUtil.writeNodeObjectId(streamId, NetworkUtil.getObjectId(self.storage))
        self.storage:writeStream(streamId, connection)
        g_server:registerObjectInStream(connection, self.storage)
    end
end

function IndividualSiloPlaceable:update(dt)
    IndividualSiloPlaceable:superClass().update(self, dt)

    if self:getCanInteract() then
        local fillLevel = 0
        local fillTypeName = self.fillLevelTextBackup

        if self.storage.currentStoredFillType ~= nil and self.storage.currentStoredFillType ~= FillType.UNKNOWN then
            local fillType = g_fillTypeManager:getFillTypeByIndex(self.storage.currentStoredFillType)
            if fillType ~= nil then
                fillTypeName = fillType.title
            end
            fillLevel = self.storage:getFillLevel(self.storage.currentStoredFillType)
        end

        g_currentMission:addExtraPrintText(fillTypeName .. ": " .. math.floor(fillLevel) .. " (" .. math.floor(100 * fillLevel / self.storage.capacityPerFillType) .. "%)")

        self:raiseActive()
    end
end

function IndividualSiloPlaceable:loadFromXMLFile(xmlFile, key, resetVehicles)
    if not IndividualSiloPlaceable:superClass().loadFromXMLFile(self, xmlFile, key, resetVehicles) then
        return false
    end

    if not self.unloadingStation:loadFromXMLFile(xmlFile, key .. ".unloadingStation") then
        return false
    end

    if self.storage == nil or not self.storage:loadFromXMLFile(xmlFile, key .. ".storage") then
        return false
    end

    return true
end

function IndividualSiloPlaceable:saveToXMLFile(xmlFile, key, usedModNames)
    IndividualSiloPlaceable:superClass().saveToXMLFile(self, xmlFile, key, usedModNames)

    self.unloadingStation:saveToXMLFile(xmlFile, key .. ".unloadingStation", usedModNames)

    if self.storage ~= nil then
        self.storage:saveToXMLFile(xmlFile, key .. ".storage", usedModNames)
    end
end

function IndividualSiloPlaceable:fillLevelUpdate(fillLevel, capacity, fillTypeIndex)
    if self.displays ~= nil then
        for i = 1, #self.displays do
            local display = self.displays[i]

            if display.type == "FILL_LEVEL" then
                self:setDisplayValue(display, math.floor(math.max(0, fillLevel)))
            elseif display.type == "PERCENT" then
                local percent = math.min(math.max(fillLevel / capacity, 0), 1)
                self:setDisplayValue(display, math.abs(percent * 100))
            end
        end
    end

    local visibility = fillLevel > 0.0001
    local updateMaterial = self.lastFillTypeIndex ~= fillTypeIndex

    if self.fillPlanes ~= nil then
        for i = 1, #self.fillPlanes do
            local fillPlane = self.fillPlanes[i]

            if fillPlane.updateTranslation then
                local x, y, z = getTranslation(fillPlane.node)
                y = math.min(fillPlane.minY + fillLevel / capacity * (fillPlane.maxY - fillPlane.minY), fillPlane.maxY)
                setTranslation(fillPlane.node, x, y, z)
            end

            if updateMaterial and fillPlane.materials ~= nil then
                local newMaterial = fillPlane.materials[fillTypeIndex]
                if newMaterial ~= nil then
                    local nodeMaterial = getMaterial(fillPlane.node, 0)
                    if nodeMaterial ~= newMaterial then
                        setMaterial(fillPlane.node, newMaterial, 0)
                    end
                else
                    local nodeMaterial = getMaterial(fillPlane.node, 0)
                    if nodeMaterial ~= fillPlane.originalMaterial then
                        setMaterial(fillPlane.node, fillPlane.originalMaterial, 0)
                    end
                end
            end

            setVisibility(fillPlane.node, visibility or fillPlane.alwaysVisible)
        end
    end

    if self.fillTypeIcons ~= nil then
        local material

        if updateMaterial then
            if fillTypeIndex ~= FillType.UNKNOWN then
                material = g_materialManager:getMaterial(fillTypeIndex, "icon", 1)
            end

            if material == nil and self.fillTypeIcons.originalMaterial ~= nil then
                material = originalMaterial
            end
        end

        if updateMaterial or self.lastVisibilty ~= visibility then
            for i = 1, #self.fillTypeIcons.icons do
                local icon = self.fillTypeIcons.icons[i]

                if material ~= nil then
                    setMaterial(icon.node, material, 0)
                    setShaderParameter(icon.node, "iconColor", icon.color[1], icon.color[2], icon.color[3], icon.color[4], false)
                end

                setVisibility(icon.node, visibility or self.fillTypeIcons.alwaysVisible)
            end
        end
    end

    self.lastVisibilty = visibility
    self.lastFillTypeIndex = fillTypeIndex
end

function IndividualSiloPlaceable:setDisplayValue(display, value)
    if display.lastvalue ~= value then
        display.lastvalue = value
        I3DUtil.setNumberShaderByValue(display.numbers, math.min(display.maxValue, value), 0, display.showOnEmpty)
    end
end

function IndividualSiloPlaceable:getCanInteract()
    if g_currentMission.controlPlayer and self.playerInRange then
        return self.storage ~= nil
    end

    if not g_currentMission.controlPlayer then
        for vehicle in pairs(self.vehiclesInRange) do
            if vehicle ~= nil and vehicle.getIsActiveForInput ~= nil then
                return self.storage ~= nil and vehicle:getIsActiveForInput(true)
            end
        end
    end

    return false
end

function IndividualSiloPlaceable:collectPickObjects(node)
    local foundNode = false

    for _, unloadTrigger in ipairs(self.unloadingStation.unloadTriggers) do
        if node == unloadTrigger.exactFillRootNode then
            foundNode = true
            break
        end
    end

    if not foundNode then
        for _, loadTrigger in ipairs(self.loadingStation.loadTriggers) do
            if node == loadTrigger.triggerNode then
                foundNode = true
                break
            end
        end
    end

    if not foundNode then
        IndividualSiloPlaceable:superClass().collectPickObjects(self, node)
    end
end

function IndividualSiloPlaceable:setOwnerFarmId(ownerFarmId, noEventSend)
    IndividualSiloPlaceable:superClass().setOwnerFarmId(self, ownerFarmId, noEventSend)

    if self.unloadingStation ~= nil then
        self.unloadingStation:setOwnerFarmId(ownerFarmId, noEventSend)
    end

    if self.loadingStation ~= nil then
        self.loadingStation:setOwnerFarmId(ownerFarmId, noEventSend)
    end

    if self.storage ~= nil then
        self.storage:setOwnerFarmId(ownerFarmId, noEventSend)
    end
end

function IndividualSiloPlaceable.loadSpecValueVolume(xmlFile, customEnvironment)
    return getXMLInt(xmlFile, "placeable.singleStorageSilo.storage#capacityPerFillType")
end

function IndividualSiloPlaceable.getSpecValueVolume(storeItem, realItem)
    if storeItem.specs.siloVolume == nil then
        return nil
    end

    return g_i18n:formatVolume(storeItem.specs.siloVolume)
end

function IndividualSiloPlaceable:onSell()
    IndividualSiloPlaceable:superClass().onSell(self)

    if self.isServer and self.totalFillTypeSellPrice > 0 then
        g_currentMission:addMoney(self.totalFillTypeSellPrice, self:getOwnerFarmId(), MoneyType.HARVEST_INCOME, true, true)
        self.totalFillTypeSellPrice = 0
    end
end

function IndividualSiloPlaceable:canBeSold()
    local totalFillLevel = 0
    local warning = g_i18n:getText("info_siloExtensionNotEmpty")

    self.totalFillTypeSellPrice = 0
    for fillTypeIndex, fillLevel in pairs(self.storage.fillLevels) do
        totalFillLevel = totalFillLevel + fillLevel

        if fillLevel > 0 then
            local lowestSellPrice = math.huge

            for _, unloadingStation in pairs (g_currentMission.storageSystem.unloadingStations) do
                if unloadingStation.isSellingPoint and unloadingStation.fillTypePrices[fillTypeIndex] ~= nil then
                    local price = unloadingStation:getEffectiveFillTypePrice(fillTypeIndex)
                    if price ~= nil and price > 0 then
                        lowestSellPrice = math.min(lowestSellPrice, price)

                        if lowestSellPrice == math.huge then
                            lowestSellPrice = IndividualSiloPlaceable.LOWEST_SELL_PRICE
                        end

                        local price = fillLevel * lowestSellPrice * IndividualSiloPlaceable.PRICE_SELL_FACTOR
                        local fillType = g_fillTypeManager:getFillTypeByIndex(fillTypeIndex)

                        self.totalFillTypeSellPrice = self.totalFillTypeSellPrice + price
                    end
                end
            end

            -- If there is no Sell Points on the map.
            if lowestSellPrice == math.huge then
                lowestSellPrice = IndividualSiloPlaceable.LOWEST_SELL_PRICE
            end

            local price = fillLevel * lowestSellPrice * SiloExtensionPlaceable.PRICE_SELL_FACTOR
            local fillType = g_fillTypeManager:getFillTypeByIndex(fillTypeIndex)
            warning = string.format("%s %s (%d %s) - %s: %s\n", warning, fillType.title, g_i18n:getFluid(fillLevel), g_i18n:getVolumeUnit(false), g_i18n:getText("ui_sellValue"), g_i18n:formatMoney(price, 0, true, true))
            self.totalFillTypeSellPrice = self.totalFillTypeSellPrice + price
        end
    end

    if totalFillLevel > 0 then
        return true, warning
    end

    return true, nil
end

function IndividualSiloPlaceable:interactionTriggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
    if onEnter or onLeave then
        if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
            if onEnter then
                self.playerInRange = true
            else
                self.playerInRange = false
            end

            self:raiseActive()
        else
            local vehicle = g_currentMission.nodeToObject[otherShapeId]
            if vehicle ~= nil then
                if onEnter then
                    if self.vehiclesInRange[vehicle] == nil then
                        self.vehiclesInRange[vehicle] = true
                        self.numVehiclesInRange = self.numVehiclesInRange + 1
                    end
                else
                    if self.vehiclesInRange[vehicle] then
                        self.vehiclesInRange[vehicle] = nil
                        self.numVehiclesInRange = self.numVehiclesInRange - 1
                    end
                end

                self:raiseActive()
            end
        end
    end
end
